#!/bin/bash
########################################################################
# build_geant4.sh
#
# Build geant4 and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_16_02.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: $*\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: $prog <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       $prog -h|--help [<product-dir>]

  Build $package and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking $prog.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking $prog.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "$@"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="$arg"; break
  done
  if [ -z "$product_dir" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> $product_dir\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "$product_dir/.upsfiles" ] && [ -w "$product_dir" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "$product_dir\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "$product_dir" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: $product_dir\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="$product_dir/ssibuildshims/$ssibuildshims_version"
  [ -f "$ssibuildshims_topdir/bin/ssi_functions" ] && \
    source "$ssibuildshims_topdir/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "$product_dir" || ssi_die "unable to setup UPS"
    setup ssibuildshims $ssibuildshims_version -z "$product_dir" || \
    ssi_die "unable to find ssibuildshims $ssibuildshims_version in
$product_dir and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "$product_dir" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  $trace verify_in_filesystem "$@" -d "$pkgdir/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "$maketar" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    $trace is_built "$@" && \
      $trace make_basic_distribution "$@" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

have_qt_build() {
  case $plat in
    x86_64) : ;;
    *) ssi_die "QT build requested for unsupported platform $plat"
  esac
  case ${OS%-$plat} in
    slf6|sl7) : ;;
    *) ssi_die "QT build requested for unsupported OS ${OS%-$plat}"
  esac
}

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. $thisdir/version_info.sh "$@" || \
  ssi_die "unable to load version info from $thisdir/version_info"
####################################

####################################
# Handle options and arguments.

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,e26,e28,c14,c15} \
  --bq-qt={,qt} \
  --bq-mt={,mt} \
  --bq-vg={,vg} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
$trace define_basics TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
$trace fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
$trace setup_cmake --latest={--ge,3.15} || exit
$trace setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
$trace maybe_make_distribution --silent

ssi_info "building $package for $OS-$plat-$qualdir (flavor $pkgflvr)"

# Ensure directories: pkgdir, sourcedir, blddir.
$trace init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

$trace cd "$sourcedir" \
  || ssi_die "unable to cd to source directory $sourcedir"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
$trace prepare_source "${ssi_args[@]}" -P "$patchdir" \
  "$tardir/$pkgtarfile" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "$DEBUG_PATCHES" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

cmake_options=(
  -DCMAKE_INSTALL_PREFIX="$pkgdir"
  -DCMAKE_INSTALL_LIBDIR=lib
  -DGEANT4_USE_SYSTEM_CLHEP:BOOL=ON
  -DCLHEP_ROOT_DIR:STRING="$CLHEP_BASE_DIR"
  -DGEANT4_USE_OPENGL_X11:BOOL=ON
  -DGEANT4_USE_GDML:BOOL=ON
  -DOpenGL_GL_PREFERENCE:STRING=LEGACY
)

(( GEANT4_DEV )) && \
  cmake_options+=(
    -DBUILD_STATIC_LIBS:BOOL=ON
  )

cflg=()
cxxflg=()

case $btype in
  prof)
    cmake_options+=(
      -DCMAKE_BUILD_TYPE:STRING=RelWithDebInfo
      -DCMAKE_CXX_FLAGS_RELWITHDEBINFO="-g -O3 -fno-omit-frame-pointer"
    )
    ;;
  debug)
    cmake_options+=(-DCMAKE_BUILD_TYPE:STRING=Debug)
    ;;
  *)
    ssi_die "accepted unknown build type $btype"
esac

case $cqual in
  e1[79]|e2[06])
    cc=gcc
    cstd=11
    cxx=g++
    cxxstd=17
    ;;
  e28)
    cc=gcc
    cstd=11
    cxx=g++
    cxxstd=20
    ;;
  c1[45])
    cc=clang
    cstd=11
    cxx=clang++
    cxxstd=17
    cxxflg+=(-stdlib=libc++)
    ;;
  *)
    ssi_die "accepted unknown qualifier $cqual"
esac
cmake_options+=(
  --debug-trycompile
  -DCMAKE_C_COMPILER=$cc
  -DCMAKE_C_STANDARD=$cstd
  -DCMAKE_CXX_COMPILER=$cxx
  -DCMAKE_CXX_STANDARD=$cxxstd
  ${cxxflg[*]:+"-DCMAKE_CXX_FLAGS=${cxxflg[*]}"}
)
if [ -n "$mt" ]; then
  cmake_options+=(
    -DGEANT4_BUILD_MULTITHREADED:BOOL=ON
    -DGEANT4_BUILD_TLS_MODEL:STRING=global-dynamic
  )
else
  cmake_options+=(-DGEANT4_BUILD_MULTITHREADED:BOOL=OFF)
fi

[ -n "$qt" ] && have_qt_build && \
  cmake_options+=(
    -DGEANT4_USE_QT:BOOL=ON
    -DQT_QMAKE_EXECUTABLE="$QT_FQ_DIR/bin/qmake"
  )

[ -n "$vg" ] && cmake_options+=(-DGEANT4_USE_USOLIDS:STRING="all")

(( want_ninja )) && cmake_options+=(-GNinja)

##################
# Configure parallelism. *C*
ncores=$(ncores)

[[ -n "$CMAKE_BUILD_PARALLEL_LEVEL" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=$ncores

# For ctest:
[[ -n "$CTEST_PARALLEL_LEVEL" ]] || \
  export CTEST_PARALLEL_LEVEL=$ncores
##################
####################################

####################################
# Configure, build, and install.
#
# Use $blddir/$pkgtardir for in-source builds, $blddir otherwise.

$trace cd "$blddir" || \
  ssi_die "unable to cd to build directory $blddir"

##################
# Configure.
$trace \
  cmake ${cmake_options[*]:+"${cmake_options[@]}"} \
  ${DEBUG_CONFIG:+--verbose} \
  "$sourcedir/$pkgtardir" \
  || ssi_die "configure failed"
##################

##################
# Build.
$trace cmake --build "$blddir" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Test.
##################

##################
# Install.
$trace cmake --install "$blddir" || ssi_die "install failed"
##################

ssi_info "finished building $package $pkgver"

####################################
# Pre-declaration processing.
config_scripts=(
  "$GEANT4_FQ_DIR/bin/geant4-config"
  "$GEANT4_FQ_DIR/bin/"*sh
  "$GEANT4_FQ_DIR/share/Geant4/geant4make/"geant4make*sh
  "$G4LIB/cmake/Geant4/"Geant4*.cmake
)

data_dirs=(
  G4NEUTRONHPDATA
  G4LEDATA
  G4LEVELGAMMADATA
  G4RADIOACTIVEDATA
  G4PARTICLEXSDATA
  G4PIIDATA
  G4REALSURFACEDATA
  G4SAIDXSDATA
  G4ABLADATA
  G4INCLDATA
  G4ENSDFSTATEDATA
)

# Preprocess config scripts to remove ... "quirks."
ssi_info "pre-post-processing ${config_scripts[@]} ..."
for dvar in "${data_dirs[@]}"; do
  eval dval="\$$dvar"
  $trace \
    sed -Ei'' \
    -e 's&\b('"$dvar"'\|)[^|]+&\1'"'\"$dval\"'"'&g' \
    -e '\&\b(export|setenv)[[:space:]]+'"$dvar"'[[:space:]=]& d' \
    "${config_scripts[@]}" || break
done || ssi_die "failed pre-post-processing ${config_scripts[@]}"

$trace fix_config_script --ignore-missing \
  =CLHEP_BASE_DIR \
  =XERCESCROOT \
  =QT_FQ_DIR \
  "${data_dirs[@]/#/=}" \
  -- \
  "${config_scripts[@]}" || ssi_die "failed post-processing"

# Quick adjustment to automatic replacement to remove unwanted sets of
# double quotes and unhelpful directives.
sed -Ei'' \
    -e '\&^set\(Geant4_DATASET_DESCRIPTIONS[[:space:]]+& '"s&(\"'|'\")&'&g" \
    -e '/^[[:space:]]*geant4_set_and_check_package_variable[[:space:]]*\([[:space:]]*EXPAT_LIBRARY/ d'
  "$G4LIB/cmake/Geant4/"Geant4*.cmake
###################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
$trace declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
$trace is_built && $trace ssi_build_completed && \
  ssi_info "$package is installed at $pkgdir" && \
  $trace maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
